import asyncio

from .. import fhem
from .. import fhem_pythonbinding as fhepy
from .. import generic

# imports for dynamical usage, do NOT remove
from .devices.gateway import Gateway  # noqa: F401
from .devices.sensor import (  # noqa: F401
    ContactSensor,
    HTSensor,
    MotionSensor,
    WaterLeakSensor,
)

device_type_mapping = {
    "lumi.sensor_magnet": "ContactSensor",
    "lumi.sensor_magnet.v2": "ContactSensor",
    "lumi.sensor_magnet.aq2": "ContactSensor",
    "lumi.sensor_wleak": "WaterLeakSensor",
    "lumi.sensor_wleak.aq1": "WaterLeakSensor",
    "lumi.sensor_ht": "HTSensor",
    "lumi.sensor_ht.v1": "HTSensor",
    "lumi.sensor_ht.v2": "HTSensor",
    "lumi.weather": "HTSensor",
    "lumi.weather.v1": "HTSensor",
    "lumi.sensor_motion": "MotionSensor",
    "lumi.sensor_motion.v1": "MotionSensor",
    "lumi.sensor_motion.v2": "MotionSensor",
    "lumi.gateway": "Gateway",
    "lumi.gateway.mgl03": "Gateway",
}


class xiaomi_gateway3_device(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self._fhempy_gateway = None
        self._fhempy_device = None
        self.loop = asyncio.get_event_loop()

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)

        if len(args) < 5:
            return (
                "Usage: define devname fhempy xiaomi_gateway3_device"
                " <GATEWAY_NAME> <DID>"
            )

        self.gw_name = args[3]
        self.did = args[4]

        hash["GATEWAY"] = self.gw_name
        hash["DID"] = self.did

        # change gateway did to 0, we just needed it for DID internals
        if self.did.find("0x") >= 0:
            self.did = "lumi.0"

        await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "offline", 1)

        self.create_async_task(self.connect_gw())

    async def connect_gw(self):
        while self._fhempy_gateway is None:
            self._fhempy_gateway = fhepy.getFhemPyDeviceByName(self.gw_name)
            if self._fhempy_gateway:
                try:
                    await self._fhempy_gateway.register_device(self, self.update)
                    await fhem.readingsSingleUpdateIfChanged(
                        self.hash, "state", "online", 1
                    )
                except Exception:
                    self._fhempy_gateway = None
                    pass
            else:
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash, "state", f"gateway {self.gw_name} not found", 1
                )
            await asyncio.sleep(10)

    async def initialize(self, device):
        if self._fhempy_gateway is None:
            return

        # first update, set attributes and device readings like model, sid, ...
        if device["model"] not in device_type_mapping:
            self.logger.error(
                f"{device['model']} not yet supported, please report an issue here: "
                f"https://github.com/dominikkarall/fhempy/issues"
            )
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "state", f"unsupported device: {device['model']}", 1
            )
            return
        # create device based on device model
        self._fhempy_device = globals()[device_type_mapping[device["model"]]](
            self.logger, self._fhempy_gateway
        )
        self._fhempy_device.set_hash(self.hash)
        await self._fhempy_device.initialize(device)
        self._fhempy_gateway.gateway3.set_entity(self._fhempy_device)
        self._fhempy_gateway.gateway3.set_stats(self._fhempy_device)

    def update(self, data):
        if self._fhempy_device is not None:
            self._fhempy_device.update(data)

    # FHEM functions which will be redirected to device type class
    async def FW_detailFn(self, hash, args, argsh):
        if self._fhempy_device is None:
            return await super().FW_detailFn(hash, args, argsh)
        return await self._fhempy_device.FW_detailFn(hash, args, argsh)

    async def Set(self, hash, args, argsh):
        if self._fhempy_device is None:
            return await super().Set(hash, args, argsh)
        return await self._fhempy_device.Set(hash, args, argsh)

    async def Attr(self, hash, args, argsh):
        if self._fhempy_device is None:
            return await super().Attr(hash, args, argsh)
        return await self._fhempy_device.Attr(hash, args, argsh)

    async def Undefine(self, hash):
        await super().Undefine(hash)
        if self._fhempy_device is not None:
            await self._fhempy_device.Undefine(hash)



import asyncio
import inspect
import json
import os
import markdown
import functools

from fhempy.lib import fhem

from . import utils


class FhemModule:
    def __init__(self, logger):
        self.logger = logger
        self.loop = asyncio.get_event_loop()
        self._tasks = []
        self._conf_set = {}
        self._conf_attr = {}
        self.readme_str = None

    def set_attr_config(self, attr_config):
        self._conf_attr = attr_config

    def set_set_config(self, set_config):
        self._conf_set = set_config

    # FHEM FUNCTION
    async def FW_detailFn(self, hash, args, argsh):
        (FW_wname, d, room, pageHash) = args
        ret = """<script type="text/javascript">
        function displayHelp() {
          var x = document.getElementById("fhempyReadme");
          if (x.style.display === "none") {
            x.style.display = "block";
          } else {
            x.style.display = "none";
          }
          var off = $("#fhempyReadme").position().top-20;
          $('body, html').animate({scrollTop:off}, 500);
        }
        
        $(document).ready(function() {
          $("#content").append('<div class="makeTable help" id="fhempyReadme"></div>');
          $("#fhempyReadme").html(`###README_HELP_STRING###`);
          document.getElementById("fhempyReadme").style.display = "none";

          var helpCmdStr = '###SET_CMD_CONFIG_STRING###';
          var helpCmdJson = JSON.parse(helpCmdStr);
          $("select.set").change(helpSetAction);
          function
          helpSetAction(){
            var cmd = $("select.set").val()
            if(helpCmdJson[cmd] && helpCmdJson[cmd].help) {
              if (document.getElementById("idCmdHelp")===null) {
                $('<div id="idCmdHelp" class="makeTable help"></div>')
                      .insertBefore("div.makeTable.internals");
              }
              $("div#idCmdHelp").html(helpCmdJson[cmd].help + "<br>");
            } else {
              $("div#idCmdHelp").remove();
            }
          }
          helpSetAction();

          var helpAttrStr = '###ATTR_CONFIG_STRING###';
          var helpAttrJson = JSON.parse(helpAttrStr);
          $("select.attr").change(helpAttrAction);
          function
          helpAttrAction(){
            var cmd = $("select.attr").val()
            if(helpAttrJson[cmd] && helpAttrJson[cmd].help) {
              if (document.getElementById("idAttrHelp")===null) {
                $('<div id="idAttrHelp" class="makeTable help"></div>')
                      .insertBefore("div.makeTable.attributes");
              }
              $("div#idAttrHelp").html(helpAttrJson[cmd].help + "<br>");
            } else {
              $("div#idAttrHelp").remove();
            }
          }
          helpAttrAction();

          var helpLink = document.getElementById("content")
            .getElementsByClassName("detLink devSpecHelp");
          helpLink[0].innerHTML = '<div class="detLink devSpecHelp"><a href="#" onclick="displayHelp();return false;">Device specific help</a></div>';
        });
        </script>"""
        js_set_conf = {}
        for key in self._conf_set:
            if "help" in self._conf_set[key]:
                js_set_conf[key] = {}
                js_set_conf[key]["help"] = self._conf_set[key]["help"]
        ret = ret.replace(
            "###SET_CMD_CONFIG_STRING###",
            json.dumps(js_set_conf),
        )
        js_attr_conf = {}
        for key in self._conf_attr:
            if "help" in self._conf_attr[key]:
                js_attr_conf[key] = {}
                js_attr_conf[key]["help"] = self._conf_attr[key]["help"]
        ret = ret.replace(
            "###ATTR_CONFIG_STRING###",
            json.dumps(js_attr_conf),
        )

        # add readme as help
        if self.readme_str is not None:
            ret = ret.replace("###README_HELP_STRING###", self.readme_str)

        return ret

    def _get_readme_content(self):
        from fhempy import lib

        initfile = inspect.getfile(lib)
        fhempy_root = os.path.dirname(initfile)
        try:
            readme_md = ""
            with open(
                fhempy_root + "/" + self.hash["FHEMPYTYPE"] + "/README.md", "r"
            ) as f:
                readme_md = f.read()

            readme_str = markdown.markdown(readme_md)
            return readme_str
            # make_html_images_inline(
            #    readme_str, fhempy_root + "/" + self.hash["FHEMPYTYPE"] + "/"
            # )

        except FileNotFoundError:
            return f"No README.md file found for {self.hash['FHEMPYTYPE']}."

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        self.hash = hash
        check_init_done = await fhem.init_done(self.hash)
        if check_init_done == 1:
            if await fhem.AttrVal(self.hash["NAME"], "room", "") == "":
                await fhem.CommandAttr(self.hash, f"{self.hash['NAME']} room fhempy")
            if await fhem.AttrVal(self.hash["NAME"], "group", "") == "":
                await fhem.CommandAttr(
                    self.hash,
                    (f"{self.hash['NAME']} group " f"{self.hash['FHEMPYTYPE']}"),
                )
        self.readme_str = await utils.run_blocking(
            functools.partial(self._get_readme_content)
        )
        await utils.handle_define_attr(self._conf_attr, self, hash)

    # FHEM FUNCTION
    async def Attr(self, hash, args, argsh):
        return await utils.handle_attr(self._conf_attr, self, hash, args, argsh)

    # FHEM FUNCTION
    async def Set(self, hash, args, argsh):
        return await utils.handle_set(self._conf_set, self, hash, args, argsh)

    def create_async_task(self, coro):
        task = asyncio.create_task(coro)
        self._tasks.append(task)
        return task

    def cancel_async_task(self, task):
        self._tasks.remove(task)
        task.cancel()

    # FHEM FUNCTION
    async def Undefine(self, hash):
        # cancel all tasks
        for task in self._tasks:
            task.cancel()



"""
A simple wrapper for bluepy's btle.Connection.
Handles Connection duties (reconnecting etc.) transparently.
"""
import codecs
import logging
import re

import dbus
from bluepy import btle

DEFAULT_TIMEOUT = 1

_LOGGER = logging.getLogger("core.ble")


class Peripheral(btle.Peripheral):
    def __init__(
        self, deviceAddr=None, addrType=btle.ADDR_TYPE_PUBLIC, iface=None, timeout=None
    ):
        btle.BluepyHelper.__init__(self)
        self._serviceMap = None  # Indexed by UUID
        (self.deviceAddr, self.addrType, self.iface) = (None, None, None)

        if isinstance(deviceAddr, btle.ScanEntry):
            self._connect(
                deviceAddr.addr, deviceAddr.addrType, deviceAddr.iface, timeout
            )
        elif deviceAddr is not None:
            self._connect(deviceAddr, addrType, iface, timeout)

    def _connect(self, addr, addrType=btle.ADDR_TYPE_PUBLIC, iface=None, timeout=None):
        if len(addr.split(":")) != 6:
            raise ValueError("Expected MAC address, got %s" % repr(addr))
        if addrType not in (btle.ADDR_TYPE_PUBLIC, btle.ADDR_TYPE_RANDOM):
            raise ValueError(
                "Expected address type public or random, got {}".format(addrType)
            )
        self._startHelper(iface)
        self.addr = addr
        self.addrType = addrType
        self.iface = iface
        if iface is not None:
            self._writeCmd("conn %s %s %s\n" % (addr, addrType, "hci" + str(iface)))
        else:
            self._writeCmd("conn %s %s\n" % (addr, addrType))
        rsp = self._getResp("stat", timeout)
        if rsp is None:
            raise btle.BTLEDisconnectError(
                "Timed out while trying to connect to peripheral %s, addr type: %s"
                % (addr, addrType),
                rsp,
            )
        while rsp["state"][0] == "tryconn":
            rsp = self._getResp("stat", timeout)
        if rsp["state"][0] != "conn":
            self._stopHelper()
            raise btle.BTLEDisconnectError(
                "Failed to connect to peripheral %s, addr type: %s" % (addr, addrType),
                rsp,
            )

    def connect(self, addr, addrType=btle.ADDR_TYPE_PUBLIC, iface=None, timeout=None):
        if isinstance(addr, btle.ScanEntry):
            self._connect(addr.addr, addr.addrType, addr.iface, timeout)
        elif addr is not None:
            self._connect(addr, addrType, iface, timeout)


class BTLEConnection(btle.DefaultDelegate):
    """Representation of a BTLE Connection."""

    def __init__(self, mac, keep_connected=False, connection_established_callback=None):
        """Initialize the connection."""
        btle.DefaultDelegate.__init__(self)

        self._ifaces = self.get_hci_ifaces()
        self._iface_idx = 0

        self._conn = None
        self._mac = mac
        self._callbacks = {}
        self._keep_connected = keep_connected
        self._connection_etablished_callback = connection_established_callback

    def set_keep_connected(self, new_state):
        self._keep_connected = new_state
        if new_state == False:
            if self._conn:
                try:
                    self._conn.disconnect()
                except Exception:
                    pass

    def next_iface(self):
        self._nr_conn_errors += 1
        self._iface_idx = (self._iface_idx + 1) % len(self._ifaces)
        if self._nr_conn_errors >= len(self._ifaces) * 5:
            return False
        return True

    def get_hci_ifaces(self):
        iface_list = []
        bus = dbus.SystemBus()
        manager = dbus.Interface(
            bus.get_object("org.bluez", "/"), "org.freedesktop.DBus.ObjectManager"
        )
        objects = manager.GetManagedObjects()
        for path, interfaces in objects.items():
            adapter = interfaces.get("org.bluez.Adapter1")
            if adapter is None:
                continue
            iface_list.append(re.search(r"\d+$", path)[0])
        return iface_list

    def __enter__(self):
        """
        Context manager __enter__ for connecting the device
        :rtype: btle.Peripheral
        :return:
        """
        conn_state = ""
        try:
            if self._conn:
                conn_state = self._conn.getState()
        except (btle.BTLEInternalError, btle.BTLEDisconnectError):
            self._conn = None

        if self._conn is None or conn_state != "conn":
            self._conn = Peripheral()
            self._conn.withDelegate(self)
            self._nr_conn_errors = 0
            _LOGGER.debug("Trying to connect to %s", self._mac)
            while True:
                # try to connect with all ifaces
                try:
                    self._conn.connect(
                        self._mac, iface=self._ifaces[self._iface_idx], timeout=10
                    )
                    if self._connection_etablished_callback is not None:
                        self._connection_etablished_callback(self._mac)
                    break
                except btle.BTLEException as ex:
                    _LOGGER.debug(
                        "Unable to connect to the device %s using iface %s, retrying: %s",
                        self._mac,
                        self._ifaces[self._iface_idx],
                        ex,
                    )
                    try:
                        self._conn.connect(
                            self._mac, iface=self._ifaces[self._iface_idx], timeout=10
                        )
                        if self._connection_etablished_callback is not None:
                            self._connection_etablished_callback(self._mac)
                        break
                    except Exception as ex2:
                        _LOGGER.debug(
                            "Second connection try to %s using ifaces %s failed: %s",
                            self._mac,
                            self._ifaces[self._iface_idx],
                            ex2,
                        )
                        if self.next_iface() is False:
                            # tried all ifaces, raise exception
                            raise

        _LOGGER.debug("Connected to %s", self._mac)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._conn and self._keep_connected is False:
            self._conn.disconnect()
            self._conn = None

    def handleNotification(self, handle, data):
        """Handle Callback from a Bluetooth (GATT) request."""
        _LOGGER.debug(
            "Got notification from %s: %s", handle, codecs.encode(data, "hex")
        )
        if handle in self._callbacks:
            for callback in self._callbacks[handle]:
                callback(data)

    @property
    def mac(self):
        """Return the MAC address of the connected device."""
        return self._mac

    def set_callback(self, handle, function):
        """Set the callback for a Notification handle. It will be called with the parameter data, which is binary."""
        if handle not in self._callbacks:
            self._callbacks[handle] = []
        self._callbacks[handle].append(function)

    def read_characteristic(self, handle):
        with self:
            return self._conn.readCharacteristic(handle)

    def write_characteristic(
        self, handle, value, timeout=DEFAULT_TIMEOUT, with_response=True
    ):
        """Write a GATT Command without callback - not utf-8."""
        try:
            with self:
                _LOGGER.debug(
                    "Writing %s to %s with with_response=%s",
                    codecs.encode(value, "hex"),
                    handle,
                    with_response,
                )
                self._conn.writeCharacteristic(
                    handle, value, withResponse=with_response
                )
                if timeout:
                    _LOGGER.debug("Waiting for notifications for %s", timeout)
                    self._conn.waitForNotifications(timeout)
        except btle.BTLEException as ex:
            _LOGGER.debug("Got exception from bluepy while making a request: %s", ex)
            raise
